@@ -27,17 +27,28 @@
 #include <net/clientexception.h>
 #include <net/socket_helper.h>
 #include <net/socket_msg.h>
+#include <net/downloadhelper.h>
 #include <core/avatarmanager.h>
+#include <core/crypthelper.h>
 #include <qttoolsinterface.h>
 
 #include <game.h>
 #include <playerinterface.h>
 
+#include "tinyxml.h"
+#include "zlib.h"
 #include <boost/bind.hpp>
+#include <boost/iostreams/filtering_streambuf.hpp>
+#include <boost/iostreams/copy.hpp>
+#include <boost/iostreams/filter/zlib.hpp>
+#include <boost/filesystem.hpp>
 
+#include <iostream>
+#include <fstream>
 #include <sstream>
 
 using namespace std;
+using namespace boost::filesystem;
 
 #define CLIENT_WAIT_TIMEOUT_MSEC	50
 #define CLIENT_CONNECT_TIMEOUT_SEC	10
@@ -92,7 +103,10 @@ ClientStateInit::Process(ClientThread &client)
 	setsockopt(context.GetSocket(), SOL_SOCKET, SO_NOSIGPIPE, (char *)&nosigpipe, sizeof(nosigpipe));
 #endif
 
-	client.SetState(ClientStateStartResolve::Instance());
+	if (context.GetUseServerList())
+		client.SetState(ClientStateStartServerListDownload::Instance());
+	else
+		client.SetState(ClientStateStartResolve::Instance());
 
 	return MSG_SOCK_INIT_DONE;
 }
@@ -117,7 +131,7 @@ ClientStateStartResolve::~ClientStateStartResolve()
 int
 ClientStateStartResolve::Process(ClientThread &client)
 {
-	int retVal;
+	int retVal = MSG_SOCK_INTERNAL_PENDING;
 
 	ClientContext &context = client.GetContext();
 
@@ -148,8 +162,6 @@ ClientStateStartResolve::Process(ClientThread &client)
 
 		ClientStateResolving::Instance().SetResolver(resolver.release());
 		client.SetState(ClientStateResolving::Instance());
-
-		retVal = MSG_SOCK_INTERNAL_PENDING;
 	}
 
 	return retVal;
@@ -223,6 +235,253 @@ ClientStateResolving::Cleanup()
 
 //-----------------------------------------------------------------------------
 
+ClientStateStartServerListDownload &
+ClientStateStartServerListDownload::Instance()
+{
+	static ClientStateStartServerListDownload state;
+	return state;
+}
+
+ClientStateStartServerListDownload::ClientStateStartServerListDownload()
+{
+}
+
+ClientStateStartServerListDownload::~ClientStateStartServerListDownload()
+{
+}
+
+int
+ClientStateStartServerListDownload::Process(ClientThread &client)
+{
+	int retVal = MSG_SOCK_INTERNAL_PENDING;
+
+	const ClientContext &context = client.GetContext();
+	path tmpServerListPath(context.GetCacheDir());
+	string serverListUrl(context.GetServerListUrl());
+	// Retrieve the file name from the URL.
+	size_t pos = serverListUrl.find_last_of('/');
+	if (serverListUrl.empty() || pos == string::npos || ++pos >= serverListUrl.length())
+	{
+		throw ClientException(__FILE__, __LINE__, ERR_SOCK_INVALID_SERVERLIST_URL, 0);
+	}
+	tmpServerListPath /= serverListUrl.substr(pos);
+	if (exists(tmpServerListPath))
+	{
+		// Download and compare md5.
+		tmpServerListPath = change_extension(tmpServerListPath, extension(tmpServerListPath) + ".md5");
+		std::auto_ptr<DownloadHelper> downloader(new DownloadHelper);
+		downloader->Init(serverListUrl + ".md5", tmpServerListPath.directory_string());
+		ClientStateSynchronizingServerList::Instance().SetDownloadHelper(downloader.release());
+		client.SetState(ClientStateSynchronizingServerList::Instance());
+	}
+	else
+	{
+		// Download server list.
+		std::auto_ptr<DownloadHelper> downloader(new DownloadHelper);
+		downloader->Init(serverListUrl, tmpServerListPath.directory_string());
+		ClientStateDownloadingServerList::Instance().SetDownloadHelper(downloader.release());
+		client.SetState(ClientStateDownloadingServerList::Instance());
+	}
+
+	return retVal;
+}
+
+//-----------------------------------------------------------------------------
+
+ClientStateSynchronizingServerList &
+ClientStateSynchronizingServerList::Instance()
+{
+	static ClientStateSynchronizingServerList state;
+	return state;
+}
+
+ClientStateSynchronizingServerList::ClientStateSynchronizingServerList()
+{
+}
+
+ClientStateSynchronizingServerList::~ClientStateSynchronizingServerList()
+{
+}
+
+void
+ClientStateSynchronizingServerList::SetDownloadHelper(DownloadHelper *helper)
+{
+	m_downloadHelper.reset(helper);
+}
+
+int
+ClientStateSynchronizingServerList::Process(ClientThread &client)
+{
+	int retVal = MSG_SOCK_INTERNAL_PENDING;
+
+	if (m_downloadHelper->Process())
+	{
+		m_downloadHelper.reset(NULL);
+		ClientContext &context = client.GetContext();
+		path md5ServerListPath(context.GetCacheDir());
+
+		// No more checking needed as this was done before.
+		md5ServerListPath /= context.GetServerListUrl().substr(context.GetServerListUrl().find_last_of('/') + 1) + ".md5";
+		path serverListPath = change_extension(md5ServerListPath, "");
+		// Compare the md5 sums.
+		string tmpMd5;
+		{
+			ifstream inFile(md5ServerListPath.directory_string().c_str(), ios_base::in);
+			if (inFile.fail())
+				throw ClientException(__FILE__, __LINE__, ERR_SOCK_OPEN_MD5_FAILED, 0);
+			inFile >> tmpMd5;
+		}
+		MD5Buf downloadedMd5;
+		if (!downloadedMd5.FromString(tmpMd5))
+			throw ClientException(__FILE__, __LINE__, ERR_SOCK_INVALID_SERVERLIST_MD5, 0);
+		MD5Buf currentMd5;
+		if (!CryptHelper::MD5Sum(serverListPath.directory_string(), currentMd5))
+			throw ClientException(__FILE__, __LINE__, ERR_SOCK_INVALID_SERVERLIST_MD5, 0);
+		if (downloadedMd5 == currentMd5)
+		{
+			// Server list is still current.
+			client.SetState(ClientStateReadingServerList::Instance());
+		}
+		else
+		{
+			// Download new server list.
+			// Paranoia check before removing the file, we do not want to delete wrong files.
+			path tmpPath(serverListPath);
+			if (path(context.GetCacheDir()) == tmpPath.remove_leaf())
+			{
+				remove(serverListPath);
+				client.SetState(ClientStateStartServerListDownload::Instance());
+			}
+			else
+				throw ClientException(__FILE__, __LINE__, ERR_SOCK_INVALID_SERVERLIST_URL, 0);
+		}
+	}
+
+	return retVal;
+}
+
+//-----------------------------------------------------------------------------
+
+ClientStateDownloadingServerList &
+ClientStateDownloadingServerList::Instance()
+{
+	static ClientStateDownloadingServerList state;
+	return state;
+}
+
+ClientStateDownloadingServerList::ClientStateDownloadingServerList()
+{
+}
+
+ClientStateDownloadingServerList::~ClientStateDownloadingServerList()
+{
+}
+
+void
+ClientStateDownloadingServerList::SetDownloadHelper(DownloadHelper *helper)
+{
+	m_downloadHelper.reset(helper);
+}
+
+int
+ClientStateDownloadingServerList::Process(ClientThread &client)
+{
+	int retVal = MSG_SOCK_INTERNAL_PENDING;
+
+	if (m_downloadHelper->Process())
+	{
+		m_downloadHelper.reset(NULL);
+		client.SetState(ClientStateReadingServerList::Instance());
+	}
+
+	return retVal;
+}
+
+//-----------------------------------------------------------------------------
+
+ClientStateReadingServerList &
+ClientStateReadingServerList::Instance()
+{
+	static ClientStateReadingServerList state;
+	return state;
+}
+
+ClientStateReadingServerList::ClientStateReadingServerList()
+{
+}
+
+ClientStateReadingServerList::~ClientStateReadingServerList()
+{
+}
+
+int
+ClientStateReadingServerList::Process(ClientThread &client)
+{
+	int retVal = MSG_SOCK_INTERNAL_PENDING;
+
+	ClientContext &context = client.GetContext();
+	path zippedServerListPath(context.GetCacheDir());
+	zippedServerListPath /= context.GetServerListUrl().substr(context.GetServerListUrl().find_last_of('/') + 1);
+	path xmlServerListPath;
+	if (extension(zippedServerListPath) == ".z")
+	{
+		xmlServerListPath = change_extension(zippedServerListPath, "");
+
+		// Unzip the file using zlib.
+		try {
+			ifstream inFile(zippedServerListPath.directory_string().c_str(), ios_base::in | ios_base::binary);
+			ofstream outFile(xmlServerListPath.directory_string().c_str(), ios_base::out);
+			boost::iostreams::filtering_streambuf<boost::iostreams::input> in;
+			in.push(boost::iostreams::zlib_decompressor());
+			in.push(inFile);
+			boost::iostreams::copy(in, outFile);
+		} catch (...)
+		{
+			throw ClientException(__FILE__, __LINE__, ERR_SOCK_UNZIP_FAILED, 0);
+		}
+	}
+	else
+		xmlServerListPath = zippedServerListPath;
+
+	// Parse the server address.
+	TiXmlDocument doc(xmlServerListPath.directory_string());
+
+	if (doc.LoadFile())
+	{
+		TiXmlHandle docHandle(&doc);	
+		const TiXmlElement *firstServer = docHandle.FirstChild("ServerList" ).FirstChild("Server").ToElement();
+		if (firstServer)
+		{
+			const TiXmlNode *addrNode;
+			if (context.GetAddrFamily() == AF_INET6)
+				addrNode = firstServer->FirstChild("IPv6Address");
+			else
+				addrNode = firstServer->FirstChild("IPv4Address");
+			const TiXmlNode *portNode = firstServer->FirstChild("Port");
+
+			if (!addrNode || !addrNode->ToElement() || !portNode || !portNode->ToElement())
+				throw ClientException(__FILE__, __LINE__, ERR_SOCK_INVALID_SERVERLIST_XML, 0);
+
+			context.SetServerAddr(addrNode->ToElement()->Attribute("value"));
+
+			int tmpPort = 0;
+			portNode->ToElement()->QueryIntAttribute("value", &tmpPort);
+			context.SetServerPort((unsigned)tmpPort);
+			retVal = MSG_SOCK_SERVER_LIST_DONE;
+		}
+		else
+			throw ClientException(__FILE__, __LINE__, ERR_SOCK_INVALID_SERVERLIST_XML, 0);
+	}
+	else
+		throw ClientException(__FILE__, __LINE__, ERR_SOCK_INVALID_SERVERLIST_XML, 0);
+
+	client.SetState(ClientStateStartResolve::Instance());
+
+	return retVal;
+}
+
+//-----------------------------------------------------------------------------
+
 ClientStateStartConnect &
 ClientStateStartConnect::Instance()
 {
@@ -413,6 +672,9 @@ AbstractClientStateReceiving::Process(ClientThread &client)
 			NetPacketRemovedFromGame::Data removedData;
 			tmpPacket->ToNetPacketRemovedFromGame()->GetData(removedData);
 			client.ClearPlayerDataList();
+			// Resubscribe Lobby messages.
+			client.ResubscribeLobbyMsg();
+			// Show Lobby.
 			client.GetCallback().SignalNetClientWaitDialog();
 			client.GetCallback().SignalNetClientRemovedFromGame(removedData.removeReason);
 			client.SetState(ClientStateWaitJoin::Instance());
@@ -754,8 +1016,12 @@ ClientStateSynchronizeStart::Process(ClientThread &client)
 
 	if (client.IsSynchronized())
 	{
+		// Acknowledge start.
 		boost::shared_ptr<NetPacket> startAck(new NetPacketStartEventAck);
 		client.GetSender().Send(client.GetContext().GetSessionData(), startAck);
+		// Unsubscribe lobby messages.
+		client.UnsubscribeLobbyMsg();
+
 		client.SetState(ClientStateWaitStart::Instance());
 	}
 
@@ -885,6 +1151,9 @@ ClientStateWaitHand::InternalProcess(ClientThread &client, boost::shared_ptr<Net
 			if (!tmpPlayer)
 				throw ClientException(__FILE__, __LINE__, ERR_NET_UNKNOWN_PLAYER_ID, 0);
 			client.GetGui().logPlayerWinGame(tmpPlayer->getMyName(), curGame->getMyGameID());
+			// Resubscribe Lobby messages.
+			client.ResubscribeLobbyMsg();
+			// Show Lobby dialog.
 			client.GetCallback().SignalNetClientWaitDialog();
 			client.SetState(ClientStateWaitGame::Instance());
 			retVal = MSG_NET_GAME_CLIENT_END;
